//// データ関係のクラス

enum PreviewMode {stop, playing, pause};
// Preview関係の再生のコントロール
class Preview {
  PreviewMode playing;
  int         count;
  boolean     resized;
  int         delay_time;       // イメージ間の遅れ時間計測用（millis()で差を得る）
  int         delay_per_image;  // イメージ間の遅れ時間（プレビュー専用）
  boolean     enable_image;     // 画像データファイルも読み込むか
  boolean     enable_scale;     // 計測時のスケールを有効化するか
  boolean     enable_stick;     // 線分表示を有効化するか
  boolean     export_enable;    // プレビューと同時にexportも実行
  int [][]    define_stick;     // Stick picture(line segment)の定義
  
  Preview() {
    playing         = PreviewMode.stop;
    resized         = false;
    enable_image    = false;
    enable_scale    = false;
    enable_stick    = false;
    count           = 0;
    export_enable   = false;
    delay_time      = millis();
    delay_per_image = Delay_preview_per_image;
    define_stick    = Define_stick;
  }
  // 画像読み込みの無効／有効切り替え
  void switch_enable_image() {
    enable_image = (enable_image ? false : true);
  }
  // スケールの無効／有効切り替え
  void switch_enable_scale() {
    enable_scale = (enable_scale ? false : true);
  }
  // 線分表示（stick picture）の無効／有効切り替え
  void switch_enable_stick() {
    enable_stick = (enable_stick ? false : true);
  }
  // プレビュー開始
  void start_preview() {
    start_preview(false);
  }
  void start_preview(boolean ex) {  // export時は ex = true
    if ((data == null) || (data.num <= 0)) return;  // データが無い
    playing       = PreviewMode.playing;
    resized       = false;
    count         = 0;
    export_enable = ex;
  }
  void stop_preview() {
    playing       = PreviewMode.stop;
    count         = 0;
    export_enable = false;
  }
  // 一つ前のコマに
  void prev_image() {
    prev_image(1);
  }
  void prev_image(int n) {
    count -= n;
    if (count < 0) count = 0;
    resized = false;
  }
  // 一つ後のコマに
  void next_image() {
    next_image(1);
  }
  void next_image(int n) {
    count += n;
    if (count >= data.num) count = data.num - 1;
    resized = false;
  }
  // 最初のコマに
  void first_image() {
    count   = 0;
    resized = false;
  }
  void last_image() {
    count   = data.num - 1;
    resized = false;
  }
  // ディレイタイマーセット
  void set_timer() {
    delay_time = millis();
  }
  // 後処理（主にディレイ）
  void post_process(Data data) {
    int past_time;

    if (! resized) return;
    if (playing == PreviewMode.stop) return;

    past_time = millis() - delay_time;
    if (count + 1 >= data.num) {    // 最後のコマの後
      if (!export_enable) {  // プレビュー
        playing = PreviewMode.pause;  // 強制的に一時停止モードに遷移
        return;
      } else {               // エクスポート
        if (past_time < Delay_export_after)  return;
        frameCount = count;
        saveFrame(ex_file.fullpathname);
        shortMessage.set_message("Exported image file : " + data.num + " files");
        stop_preview();
      }
    } else {                        // コマ間
      if (! export_enable) {  // プレビュー
        if (keyStatus.shift) {
          if (past_time < delay_per_image * 5)     return;  // SHIFTを押すと速度を1/5に減速
        } else {
          if (past_time < delay_per_image)         return;
        }
        if (playing == PreviewMode.pause)        return;
      } else {                // エクスポート
        if (past_time < Delay_export_per_image)  return;
        frameCount = count;
        saveFrame(ex_file.fullpathname);
      }
      next_image();
    }
    return;
  }
  void readDefineStick(String fname) {
    File      fp;
    String [] lines;
    int [][]  def_stick = {};

    if ((fp = new File(fname)) == null) return;
    if ((lines = loadStrings(fname)) == null) {
      shortMessage.set_message("Can't read define_stick file");
      return;
    }
    // 書式のチェック
    for (int i = 0; i < lines.length; i++) {
      CsvStrings csv = new CsvStrings(lines[i]);
      for (int j = 0; j < 6; j++) {
        if (!csv.next() || ! csv.is_float() || int(csv.value) < 0)  {  // start, end, R, G, B, Width
          shortMessage.set_message("Error: format error (not integer or negative)");
          return;
        }
      }
      if (csv.next()) {
        shortMessage.set_message("Error: format error (too much, 6 parameters)");
        return;  // 7項目以上あったらエラー
      }
    }
    // 読み込み
    for (int i = 0; i < lines.length; i++) {
      int [] onedef = {0, 0, 0, 0, 0, 0};
      CsvStrings csv = new CsvStrings(lines[i]);
      for (int j = 0; j < 6; j++) {
        csv.next();
        onedef[j] = int(csv.value);
      }
      if ((onedef[2] > 255) || (onedef[3] > 255) || (onedef[4] > 255)) {
        shortMessage.set_message("Error: format error (color code must be 0 - 255)");
        return;
      }
      def_stick = (int [][])append(def_stick, onedef);
    }
    define_stick = def_stick; //<>//
    shortMessage.set_message("Imported define_stick file :" + define_stick.length + " data");
    enable_stick = true;
  }
}

// 計測データの可視化に関するクラス
class CurrentImage {
  PImage  img;
  String  fullpathname;            // 開いている画像ファイルのフルパス名
  Preview preview;
  int     bg_color;
  
  int      org_width, org_height;  // 画像のオリジナルサイズ
  File[]  files;                   // フォルダーモード：ファイルリスト
  int     f_no;                    // フォルダーモード時の読みだし位置
  int     filter_no;               // -1 : なし，0 : GRAY, 1 : INVERT
  
  Plot  plotMark;
  
  // コンストラクター
  CurrentImage() {
    img          = null;
    fullpathname = null;
    preview      = new Preview();
    bg_color     = 0;
    org_width = org_height = -1;
    
    // プロットの初期化
    plotMark = new Plot();
  }
  // １データ１画像を作成
  void renderer(Data data) {
    File   fp;
    PImage img0;
    int    h, w;
    float  scale;
    int    index = preview.count;
    if (data.body[index] == null) return;

    // 元データのサイズを保存
    org_width  = w = data.body[index].org_width;
    org_height = h = data.body[index].org_height;
    scale          = data.body[index].scale;
    // もしscaleが有効ならばサイズを変更
    if (preview.enable_scale && scale != 1.0) {
      w = (int)(org_width  * scale);
      h = (int)(org_height * scale); 
    }
    if (! preview.resized) {  // 1サイクル目：スクリーンサイズを変更
      // 画像も読み込む場合
      if (preview.enable_image) {
        fp = new File("");
        String separator = fp.separator;
        fp = new File(data.src_fullpath + separator + data.body[index].filename);
        img0 = null;
        if (fp.isFile()) {
          if ((img0 = loadImage(fp.getAbsolutePath())) != null) {
            // 読み込み完了
            if (preview.enable_scale && scale != 1.0) img0.resize(w, h);  // 画像を拡大する場合
          }
        }
        img = img0;
      }
      if (ci.preview.export_enable) surface.setSize(w, h);
      else                          surface.setSize(w, h + 40);
      background(BG_color[bg_color][0], BG_color[bg_color][1], BG_color[bg_color][2]);
      preview.resized = true;
      preview.set_timer();
      // 画面下のショートメッセージ領域をクリア
      if (! ci.preview.export_enable) {
        fill(72, 72, 72);
        rect(0, height - 40, width, 40);
        shortMessage.set_message(ci.preview.count + ":" + data.body[index].filename, true);
      }
  } else {                  // ２サイクル目以降：プロットの描画
      // 画面下のショートメッセージ領域をクリア
      if (! ci.preview.export_enable) {
        fill(72, 72, 72);
        rect(0, height - 40, width, 40);
        shortMessage.set_message(ci.preview.count + ":" + data.body[index].filename, true);
      }
      // 画像も読み込む場合
      if (preview.enable_image) {
        if (img != null) image(img, 0, 0);
      }
      // 線分
      if (ci.preview.enable_stick) {
        int x0, y0, x1, y1;
        if (preview.define_stick.length == 0) {
          stroke(plotMark.red(), plotMark.green(), plotMark.blue()); //<>//
          strokeWeight(ci.plotMark.lineWidth);
          for (int i = 0; i < data.body[index].num - 1; i++) {
            x0 = data.body[index].px[i]; y0 = data.body[index].py[i]; x1 = data.body[index].px[i+1]; y1 = data.body[index].py[i+1];
            if ((x0 == -1) || (y1 == -1) || (x1 == -1) || (y1 == -1)) continue;  // スキップされた計測点への対応
            line(x0, y0, x1, y1);
          }
        }else {
          for (int i = 0; i < preview.define_stick.length; i++) { //<>//
            int start, end, r, g, b, weight;
            start  = preview.define_stick[i][0];
            end    = preview.define_stick[i][1];
            r      = preview.define_stick[i][2];
            g      = preview.define_stick[i][3];
            b      = preview.define_stick[i][4];
            weight = preview.define_stick[i][5];
            if ((start >= data.body[index].num) || (end >= data.body[index].num)) continue; //<>//

            x0 = data.body[index].px[start];
            y0 = data.body[index].py[start];
            x1 = data.body[index].px[end];
            y1 = data.body[index].py[end];
            if ((x0 == -1) || (y1 == -1) || (x1 == -1) || (y1 == -1)) continue;  // スキップされた計測点への対応

            stroke(r, g, b);
            strokeWeight(weight);
            if (preview.enable_scale && (scale != 1.0)) {
                   plotMark.stick(x0, y0, x1, y1, scale);
            } else plotMark.stick(x0, y0, x1, y1, 1.0);
          }
        }
        strokeWeight(1);
      }
      // プロット
      strokeWeight(ci.plotMark.lineWidth);
      for (int i = 0; i < data.body[index].num; i++) {
        int x = data.body[index].px[i];
        int y = data.body[index].py[i];
        if ((x == -1) || (y == -1)) continue; // スキップされた計測点への対応

        if (preview.enable_scale && (scale != 1.0)) {
          x = (int)(x * scale);
          y = (int)(y * scale);
        }
        plotMark.plot(x, y);
      }
      strokeWeight(1);
    }
    return;
  }
  // プレビュー
  void play(Data data) {
    if (preview.playing != PreviewMode.stop) {
      renderer(data);
      preview.post_process(data);
    }
  }
  // 背景色の変更
  void change_bg() {
    if (++bg_color >= BG_color.length) bg_color = 0; 
  }
}

// 結合されたデータのバッファー
class Data {
  OneData[] body = {};
  String    src_fullpath;  // ファイル名抜きの絶対パス（getParent()）
  private 
  int       num;       // 読み込まれた行数
  // コンストラクター 配列は確保するが，num はstore()ごとに増やしていく
  Data() {
    num          = 0;
    src_fullpath = "";
  }
  // リストを全て削除する
  void allclear() {
    body         = null;
    src_fullpath = "";
    num          = 0;
  }
  // １ファイルのデータをメモリにストア（リストを追加）する
  boolean store(OneData src) {
    OneData d = new OneData();
    body = (OneData [])append(body, d);
    num++;
    return store(body.length - 1, src);
  }
  boolean store(int index, OneData src) {
    if (index > num) return false;
    body[index].filename   = src.filename;
    body[index].org_width  = src.org_width;
    body[index].org_height = src.org_height;
    body[index].scale      = src.scale;
    body[index].num        = src.num;
    for (int i = 0; i < src.px.length; i++) {
      body[index].px = (int [])append(body[index].px, src.px[i]);
      body[index].py = (int [])append(body[index].py, src.py[i]);
    }
    return true;
  }
  // 読み込んだデータの全リストを表示（コンソール部）：デバッグ用
  void show_data_list() {
    if (data.num <= 0) return;
    for (int i = 0; i < data.num; i++) {
      println(i + ":" + body[i].filename + " " + body[i].num + "data");
    }
  }
}

// １ファイル分のデータ
class OneData {  // リスト構造
  String filename;
  int    org_width, org_height;
  float  scale;
  int    num;
  int [] px = {};
  int [] py = {};
  
  // コンストラクター
  OneData() {
    filename   = null;
    org_width  = 0; 
    org_height = 0;
    scale      = 0.0;
  }
  // 座標データを追加する
  boolean append_xy(int x, int y) {
    px = (int [])append(px, x);
    py = (int [])append(py, y);
    if ((px == null) || (py == null)) return false;
    num++;
    return true;
  }
}

//// ファイル関係のクラス
// CSV（タブ区切りの整数）のテキストを整数値に切り分けるクラス
class CsvStrings {
  String text;
  String str;
  float  value;
  // コンストラクター
  CsvStrings(String text) {
    this.set(text);
  }
  void set(String text) {
    this.text = text + "\t";
    value     =  float("NaN");  // floatの形式ではないのでNaNになる
  }
  // 次の値を取得する：正常に終了したらtrueを返し，valueに値が設定される
  boolean next() {
    if (text.indexOf("\t") == -1) return false;
    if (text.length() == 0)       return false;
    if (text.equals("\t"))        return false;
    str   = text.substring(0, text.indexOf("\t"));
    value = float(str);
    text  = text.substring(text.indexOf("\t") + 1);
    return true;
  }
  boolean is_float() {
    if (value != value) return false;  // NaN, ちょっとトリッキーな方法
    return true;
  }
}

// 指定されたデータファイルから１画像分のデータを取得してメモリーに保存するクラス
class ImportFile {
  String fullpathname;
  
  // コンストラクター
  ImportFile() {
    fullpathname = null;
  }

  // ファイル形式のチェック
  boolean validation(String fname) {
    File         fp;
    String[]     lines;  // データファイルから読み出した全テキストデータ
    int          num;    // X,Y座標のデータの組数

    // 読み込めるか？
    if ((lines = loadStrings(fname)) == null) return false; //<>//
    // １行はファイル名（ノーチェック），scale, org_width, org_height，num, X座標，Y座標，X座標，Y座標…の繰り返しか？
    for (int i = 0; i < lines.length; i++ ) {
      CsvStrings csv = new CsvStrings(lines[i]);
      csv.next();  // ファイル名
      csv.next();  // scale or 計測日時（Ver. 2.1から）
      if (csv.str.length() == 4+2+2+2+2+2) csv.next();  // Ver.2.1以降の計測日時を読み飛ばす
      if (! csv.is_float() || csv.value <= 0.0)     return false;  // scale (float)
      csv.next();
      if (! csv.is_float() || int(csv.value) <= 0)  return false;  // org_width (int)
      csv.next();
      if (! csv.is_float() || int(csv.value) <= 0)  return false;  // org_height (int)
      csv.next();
      if (! csv.is_float() || int(csv.value) <= 0)  return false;  // num (int)
      num = int(csv.value);
      // X座標，Y座標の繰り返しのチェック
      for (int j = 0; j < num; j++) {
        if (! csv.next())                             return false;                              
        if (! csv.is_float() || int(csv.value) < -1)  return false;  // X座標 (int), -1: skipped measuring
        if (! csv.next())                             return false;                              
        if (! csv.is_float() || int(csv.value) < -1)  return false;  // Y座標 (int), -1: skipped measuring
      }
    }
    return true;
  }
  
  // 実際に統合された出たファイルを読み込む（失敗時はnullを返す）
  boolean read(String fname) {
    CsvStrings   csv;
    String[]     lines;
    OneData      onedata;
    int          num;

    // ファイルチェック
    if (! validation(fname))                  return false;  //<>//
    // ファイル読み込み
    if ((lines = loadStrings(fname)) == null) return false; //<>//

    fullpathname = fname;  // 後で画像をインポートする時にパスを使うかも知れない
    // 既に読み込み済みの場合はデータ領域をクリアする．
    if (data.num != 0) data.allclear();
    
    data = new Data(); //<>//

    // 各行のデータを読み込む
    for (int i = 0; i < lines.length; i++) {
      // filename, scale, org_width, org_heightの読み出し
      onedata = new OneData();
      csv     = new CsvStrings(lines[i]);
      csv.next();
      onedata.filename   = csv.str;          // ファイル名
      csv.next();
      if (csv.str.length() == 4+2+2+2+2+2) csv.next();  // Ver.2.1以降の計測日時を読み飛ばす
      onedata.scale      = csv.value;        // scale
      csv.next();
      onedata.org_width  = int(csv.value);   // org_width
      csv.next();
      onedata.org_height = int(csv.value);   // org_height
      csv.next();
      num                = int(csv.value);   // num
  
      // X,Y座標データの読み込み
      int x, y;
      for (int j = 0; j < num; j++) {
        csv.next(); x = int(csv.value); // X座標
        csv.next(); y = int(csv.value); // Y座標
        onedata.append_xy(x, y);
      }
      // 一行分のデータをメモリーに格納（複製）
      if (! data.store(onedata)) {
        shortMessage.set_message("Error: データのメモリへの格納に失敗しました");
        fullpathname = null;
        return false;
      }
    }
    // エクスポート関係
    File fp = new File(fullpathname);
    data.src_fullpath = fp.getParent();  // データファイルの格納されているフォルダーをセット
    if (ex_file.export_folder == null) ex_file.set_export_information(data.src_fullpath); // エクスポートフォルダー未指定時
    // 連番ファイル名のフォーマット用
    ex_file.order10 = int(log(data.num) / log(10)) + 1;

    shortMessage.set_message("Data imported: " + fullpathname + " " + data.num + " lines");
    return true;
  }     
}

// 作成した静止画をファイルに出力するクラス
class ExportFile {
  String      export_folder;  // 動画データを保存する絶対パス
  int         order10;        // 連番画像ファイルの桁数
  String      fullpathname;   // 最終的なフルパス名（saveFrame（）用）
  
  // コンストラクター
  ExportFile() {
    export_folder = null;
    order10       = 4;  // #### : 0001 - 9999．ファイル読み込み時に自動調整される
    set_export_information(null);
  }
  void set_export_information(String ex_folder) {
    if (order10 < 3) order10 = 3;
    String format = "";
    for (int i = 0; i < order10; i++) format += "#";
    export_folder = ex_folder;
    File f = new File(export_folder, Image_filename + format + Image_type);
    fullpathname = f.getAbsolutePath(); 
  }
  void start_export() {
    if (data.num == 0) {
      shortMessage.set_message("Export aborted : No data");
      return;
    }
    shortMessage.clear();
    ci.preview.start_preview(true);
  }
}

// プロットに関するクラス
final String [] plotName = {"none", "cross", "circle", "circlefill", "rect", "rectfill", "scope", "scopefill"}; 

class Plot {
  char mode;       // (0: none), 1: cross,  2: circle, 3: circlefill, 4: rect, 5: rectfill, 6, scope, 7. scopefill
  int  size;
  int  p_color;    // 0, 1, 2
  int  lineWidth;

  Plot() {
    mode      = 1;  // default: cross
    size      = Plot_size;
    lineWidth = 1;
  }
  // プロットの形状を変更
  void change_shape() {
    if (++mode >= plotName.length) mode = 1;
  }
  String shapeStr() {
    return plotName[mode];
  }
  // プロットの色を変更
  void change_color() {
    if (++p_color > Plot_color.length - 1) p_color = 0; 
  }
  // プロットを描く
  void plot(int x, int y) {
    stroke(this.red(), this.green(), this.blue());
    noFill();
    switch (mode) {
      case 0 : break;
      case 1 : // cross
        cross(x, y);
        break;
      case 3 : // circlefill
        fill(this.red(), this.green(), this.blue());
      case 2 : // circle 
        circle(x, y);
        break;
      case 5 : // rectfill
        fill(this.red(), this.green(), this.blue());
      case 4 : // rect 
        rectangle(x, y); 
        break;
      case 7 : // scopefill
        fill(this.red(), this.green(), this.blue());
      case 6 : // scope      
        cross(x, y, true);
        circle(x, y);
        break;
      default : mode = 0;
    }
  }
  void cross(int x, int y) {
    cross(x, y, false);
  }
  void cross(int x, int y, boolean large) {
    int s = size;
    if (large) s *= 2;
    line(x - s, y, x + s, y);
    line(x, y - s, x, y + s);
  }
  void circle(int x, int y) {
    ellipse(x, y, size * 2, size * 2);
  }
  void rectangle(int x, int y) {
    rectMode(CENTER); rect(x, y, size, size);
  }
  // 線分を描く（太さとカラーは事前に指定すること）
  void stick(int x0, int y0, int x1, int y1, float scale) {
    if (scale != 1.0) {
      x0 = (int)(x0 * scale);
      x1 = (int)(x1 * scale);
      y0 = (int)(y0 * scale);
      y1 = (int)(y1 * scale);
    }
    noFill();
    line(x0, y0, x1, y1);
  }
  int red() {
    return Plot_color[p_color][0];
  }
  int green() {
    return Plot_color[p_color][1];
  }
  int blue() {
    return Plot_color[p_color][2];
  }
  String color_str() {
    return color2str(this.red(), this.green(), this.blue());
  }
}

String color2str(int r, int g, int b) {
  String s;
  s = "[" + r + "," + g + "," + b + "]";
  return s;
}

// キーの同時押しに対応するためのクラス
class KeyStatus {
  boolean left, right, shift, ctrl;
  KeyStatus() {
    left = right = shift = ctrl = false;
  }
  void push(int kc) {
    switch (kc) {
      case LEFT    : left  = true; break;
      case RIGHT   : right = true; break;
      case SHIFT   : shift = true; break;
      case CONTROL : ctrl  = true; break;
    }
  }
  void release(int kc) {
    switch (kc) {
      case LEFT    : left  = false; break;
      case RIGHT   : right = false; break;
      case SHIFT   : shift = false; break;
      case CONTROL : ctrl = false; break;
    }
  }
};

// 正の整数をキーボードから入力するためのクラス
class InputInteger {
  boolean inputing_now;
  int     value;
  
  InputInteger() {
    inputing_now = false;
    value        = 0;
  }
  void start() {
    start(0);
  }
  void start(int init) {
    inputing_now = true;
    value        = init;
  }
  void keyin(char c) {
    int code;  // キャラクターコード
    code = int(c) - 48;  // 0 - 9
    if ((code >= 0) && (code <= 9)) {
      add(code);
    } else if (c == BACKSPACE) {
      del();
    } else if (c == ENTER) {
      end();
    } else if (c == ' ') {
      value = 0;
    }
  }
  void end() {
    inputing_now = false;
  }
  void add(int n) {
    value *= 10;
    value += n;
  }
  void del() {
    value /= 10;
  }
}

// ウインドウとコンソールエリアに短いメッセージを一時的に表示するためのクラス
class ShortMessage {
private  int    start;
private  String message;
  ShortMessage() {
    start   = millis();
    message = "";
  }
  void clear() {
    message = "";
  }
  void set_message(String str) {
    set_message(str, false);
  }
  void set_message(String str, boolean noconsole) {
    message = str;
    if (! noconsole) println(str);
    start   = millis();
  }
  void display_short_message(int x, int y) {
    if (millis() - start < Delay_message) {
      fill(200, 200, 200);
      textSize(16);
      text(message, x, y);
    }
  }
}
